@@ -7,7 +7,7 @@ use crate::ir::base::{
     PrimitiveSymbol, PrimitiveUndefined, PrimitiveVoid, Tuple as TupleIR, Type as TypeIR,
     TypeIdent as TypeIdentIR, TypeInfo as TypeInfoIR, TypeName as TypeNameIR,
     TypeParamConfig as TypeParamConfigIR, TypeQuery as TypeQueryIR, TypeRef as TypeRefIR,
-    Union as UnionIR,
+    Union as UnionIR, WebSysBuiltin as WebSysBuiltinIR,
 };
 pub use crate::ir::base::{EnumValue, NamespaceImport};
 use enum_to_enum::WithEffects;
@@ -286,6 +286,7 @@ impl From<Namespaced<TypeInfoIR>> for EffectContainer<FlattenedTypeInfo> {
                 TypeInfoIR::LitNumber(v) => tuple_match_convert!(ns, Ref(v)),
                 TypeInfoIR::LitString(v) => tuple_match_convert!(ns, Ref(v)),
                 TypeInfoIR::LitBoolean(v) => tuple_match_convert!(ns, Ref(v)),
+                TypeInfoIR::WebSysBuiltin(b) => tuple_match_convert!(ns, Ref(b)),
                 TypeInfoIR::Ref(v) => tuple_match_convert!(ns, Ref(v)),
                 TypeInfoIR::Array { item_type } => struct_match_convert!(ns, Array { item_type }),
                 TypeInfoIR::Tuple(v) => tuple_match_convert!(ns, Tuple(v)),
@@ -615,6 +616,7 @@ impl From<Namespaced<TypeInfoIR>> for EffectContainer<TypeRef> {
             TypeInfoIR::PrimitiveNull(p) => ns.in_ns(p).into(),
             TypeInfoIR::BuiltinPromise(b) => ns.in_ns(b).into(),
             TypeInfoIR::BuiltinDate(b) => ns.in_ns(b).into(),
+            TypeInfoIR::WebSysBuiltin(b) => ns.in_ns(b).into(),
             TypeInfoIR::Array { item_type } => {
                 let item_type: EffectContainer<TypeRef> = ns.in_ns(*item_type).into();
                 combine_effects!(
@@ -823,6 +825,7 @@ pub enum Builtin {
     Map,
     Optional,
     Variadic,
+    Named(String),
 }
 
 type_ref_from_prims!(
@@ -843,6 +846,18 @@ type_ref_from_prims!(
     BuiltinPromise => Promise,
 );
 
+impl From<Namespaced<WebSysBuiltinIR>> for EffectContainer<TypeRef> {
+    fn from(src: Namespaced<WebSysBuiltinIR>) -> EffectContainer<TypeRef> {
+        EffectContainer::new(
+            TypeRef {
+                referent: TypeIdent::Builtin(Builtin::Named(format!("web_sys::{}", src.value.0))),
+                type_params: Default::default(),
+            },
+            Default::default(),
+        )
+    }
+}
+
 impl From<Namespaced<Box<TypeInfoIR>>> for EffectContainer<Box<FlattenedTypeInfo>> {
     fn from(src: Namespaced<Box<TypeInfoIR>>) -> EffectContainer<Box<FlattenedTypeInfo>> {
         src.map(|v, ns| {
@@ -1452,6 +1467,60 @@ pub fn flatten_types<Ts: IntoIterator<Item = TypeIR>>(types: Ts) -> impl Iterato
                 })
                 .unzip();
             effs.into_iter()
-                .chain(std::iter::once(v.apply_names(&names_by_id)))
+                .chain(std::iter::once(v))
+                .map(move |t| t.apply_names(&names_by_id))
         })
 }
+
+#[cfg(test)]
+mod test {
+    use super::{flatten_types, FlatType, TypeIdent};
+    use crate::fs::MemFs;
+    use crate::parse::TsTypes;
+    use crate::ArcFs;
+    use std::collections::HashMap;
+    use std::error::Error;
+    use std::path::Path;
+    use std::sync::Arc;
+
+    fn get_types_for_code(ts_code: &str) -> Result<HashMap<TypeIdent, FlatType>, Box<dyn Error>> {
+        let test_path: &Path = Path::new("/test.d.ts");
+        let mut fs: MemFs = Default::default();
+        fs.set_cwd(Path::new("/"));
+        fs.add_file_at(test_path, ts_code.to_string());
+
+        let types_by_name = TsTypes::parse(Arc::new(fs) as ArcFs, &test_path.to_string_lossy())?
+            .iter()
+            .find_map(|(path, types_by_name)| {
+                if path == test_path {
+                    Some(
+                        flatten_types(types_by_name.values().cloned())
+                            .map(|t| (t.name.clone(), t))
+                            .collect::<HashMap<_, _>>(),
+                    )
+                } else {
+                    None
+                }
+            });
+
+        assert!(types_by_name.is_some());
+
+        Ok(types_by_name.unwrap())
+    }
+
+    #[test]
+    fn test_nested_generated_names() {
+        // a tuple as a param needs to be flattened inside of a union that gets flattened
+        let code = r#"
+            export type MyType = number | ((a: [string]) => string);
+        "#;
+
+        let types = get_types_for_code(code).unwrap();
+        let debug = format!("{:?}", types);
+
+        // TODO: this is a wildly hacky way to check this but we want to
+        // make sure that nested GeneratedNames are properly transformed
+        // into real names. this avoids having to write some generic mapper.
+        assert!(!debug.contains("GeneratedName"));
+    }
+}
